;;; ffwinio.asm  FreeForth Windows I/O
;;; $Id: ffwinio.asm,v 1.1 2006/05/29 01:55:35 lavarenn Exp $

;;; --------------------------------------------------
;;; FreeForth interface to Windows file-I/O

DATA "StdIn",StdIn,0
DATA "StdOut",StdOut,0
iobuf   dd 0

CODE "exit",_exit               ; n --
        push ebx                ; process-return value
        call [ExitProcess]      ; never returns

CODE "close",_close             ; fh -- ?
        push ebp
        push eax
        push edx
        push ebx                ; HANDLE hObject
        call [CloseHandle]      ; returns 0:failed
        xor ebx,ebx             ; -- 0:ok
        or eax,eax              ; 0:failed
        jnz @f
        call [GetLastError]     ; returns error code
        mov ebx,eax             ; -- errorCode
@@:     pop edx
        pop eax
        pop ebp
        ret

;;; openr opens existing file for read-only
CODE "openr",_openr             ; @ # -- fh
        xor ecx,ecx
        inc ecx
        jmp _open

;;; openw opens (and creates if non-existing) file for read-write
CODE "openw",_openw             ; @ # -- fh
        xor ecx,ecx
_open:  mov byte[edx+ebx],0     ; append zero-terminator (another whitespace)
        push ebp                ; save ebp and eax
        push eax
        pushd 0                 ; HANDLE hTemplateFile
        pushd $80               ; DWORD FLAG_ATTRIBUTE_NORMAL
        pushd 2                 ; DWORD OPEN_ALWAYS=2     OPEN_EXISTING=3
        add [esp],ecx
        pushd 0                 ; LPSECURITY_ATTRIBUTES
        pushd ecx               ; DWORD FILE_SHARE_NONE=0 FILE_SHARE_READ=1
        pushd $C0000000         ; DWORD GENERIC_READ=8... GENERIC_WRITE=4...
        shl ecx,30
        sub [esp],ecx
        push edx                ; LPCTSTR lpFileName
        call [CreateFile]       ; returns HANDLE, -1:failed
        mov edx,eax             ; -- handle * ; negative!
        inc eax
        jnz @f
        call [GetLastError]
        mov edx,eax             ; -- error * ; positive!
@@:     pop eax                 ; restore eax and ebp
        pop ebp
        jmp drop1               ; -- handle|error
        
VECT "type",_type               ; @ # -- n ; StdOut@ write drop ;
        DUP1 [StdOut]
        call _write
        jmp drop1

CODE "write",_write             ; @ # fh -- n
        mov ecx,[WriteFile]     ; (hFile, lpBuffer, nNumberOfBytesToWrite,
        jmp _rw                 ;  lpNumberOfBytesWritten, lpOverlapped)

VECT "accept",_accept           ; @ # -- n ; StdIn@ read ;
        DUP1 [StdIn]
;;;     jmp _read
CODE "read",_read               ; @ # fh -- n
        mov ecx,[ReadFile]      ; (hFile, lpBuffer, nNumberOfBytesToRead,
_rw:    push ebp                ;  lpNumberOfBytesRead, lpOverlapped)
        push eax
        pushd 0                 ; lpOverlapped
        pushd iobuf             ; lpNumberOfBytesRead/Written
        pushd edx               ; nNumberOfBytesToRead/Write
        pushd [eax]             ; lpBuffer
        pushd ebx               ; hFile
        call ecx                ; returns 0:failed
        mov ebx,[iobuf]
        pop eax
        mov edx,[eax+4]
        lea eax,[eax+8]
        pop ebp
        ret

;;; -----------------------------------------------------
;;; FreeForth interface to Windows dynamic-link libraries

CODE "lib:`",_libdef            ; @ # <name> --
        call _colon             ; edi = hfa
        mov byte[edi+4],1       ; make it a constant
        mov [anon],ebp          ; restart anonymous
        push eax                ; save eax, ebp, edi
        push ebp
        push edi
        mov byte[edx+ebx],0     ; append zero-terminator
        pushd edx               ; LPCTSTR lpLibFileName
        call [LoadLibrary]      ; returns HINSTANCE; eax=0 on error
        pop edi                 ; restore hfa
        mov [edi],eax           ; store libhandle
        pop ebp                 ; restore ebp
        or eax,eax
        pop eax                 ; restore eax
        jnz drop2               ; --
        call _error
        CDB "lib:???"

CODE "fun:`",_fundef            ; libhandle #args <name> --
        call _colon
        ;; compile inline a harness to call library function:
        HDD $A3529455           ; 55(push ebp)94(xchg eax,esp)52(push edx)
        HDD bssend-4,4          ; A3(mov[],eax)    save callSP at bssend-4
        HDW $E853,8             ; 53(push ebx)E8(call)
        HDD -14,10              ; call offset at ebp+10 = -(ebp+14) = call 0
        sub [ebp+10],ebp
        HDD $A15AC389,14        ; 89C3(mov ebx,eax)5A(pop edx)A1(mov eax,[])
        HDD bssend-4,18         ; restore callSP from bssend-4
        HDD $C35D94,22          ; 94(xchg eax,esp)5D(pop ebp)C3(ret)
        ;; harness length = 25 bytes
        push eax                ; save eax and ebp
        push ebp
        add edi,6               ; edi points on zero-terminated hfa string
        pushd edi               ; LPCSTR lpProcName
        pushd edx               ; HMODULE libhandle
        call [GetProcAddress]   ; returns FARPROC; eax=0 on error
        pop ebp                 ; restore ebp
        add [ebp+10],eax        ; call-offset at ebp+10
        add ebp,25
        mov [anon],ebp          ; restart anonymous
        or eax,eax              ; dlsym returns null function handle on error
        pop eax                 ; restore eax
        jnz drop2               ; --
        call _error
        CDB "fun:???"

;;; -------------------------------------------------------
;;; startup code:

start:
        pushd -10               ; STD_INPUT_HANDLE
        call [GetStdHandle]
        mov [StdIn],eax
        pushd -11               ; STD_OUTPUT_HANDLE
        call [GetStdHandle]
        mov [StdOut],eax
        call [GetCommandLine]   ; eax points on command-line
        mov edx,eax
        mov ebx,edx
@@:     inc ebx
        cmp byte[ebx],0         ; command-line zero-terminated
        jnz @b
        jmp ffboot              ; args argsend --

;;; -------------------------------------------------------
;;; windows-specific linkage stuff:

;include "INCLUDE/MACRO/IMPORT32.INC"
;;; inlined here after simplification:
; ------------------------------------
macro library [name,string] {
  forward
    local _label
    dd RVA name#.lookup,0,0,RVA _label,RVA name#.address
  common
    dd 0,0,0,0,0
  forward
    _label db string,0
    rb RVA $ and 1
}
macro import name,[label,string] {
  common
    name#.lookup:
  forward
    local _label
    dd RVA _label
  common
    dd 0
    name#.address:
  forward
    label dd RVA _label
  common
    dd 0
  forward
    _label dw 0
    db string,0
    rb RVA $ and 1
}
; ------------------------------------

data import
  library kernel,'KERNEL32.DLL'
  import kernel,\
    GetStdHandle,'GetStdHandle',\
    GetCommandLine,'GetCommandLineA',\
    CreateFile,'CreateFileA',\
    ReadFile,'ReadFile',\
    WriteFile,'WriteFile',\
    CloseHandle,'CloseHandle',\
    LoadLibrary,'LoadLibraryA',\
    GetProcAddress,'GetProcAddress',\
    GetLastError,'GetLastError',\
    ExitProcess,'ExitProcess'
end data

;;; That's all folks!!
